/**
 * This file is managed by Rollup and bundles all the CodeMirror dependencies
 * into the single file resources/lib/codemirror6.bundle.dist.js.
 */
import '@codemirror/commands';
import '@codemirror/language';
import '@codemirror/search';
import '@codemirror/state';
import '@codemirror/view';
import '@lezer/highlight';

/* eslint-disable es-x/no-export-ns-from */
export * from '@codemirror/commands';
export * from '@codemirror/language';
export * from '@codemirror/search';
export * from '@codemirror/state';
export * from '@codemirror/view';
export * from '@lezer/highlight';
