.ve-init-mw-desktopArticleTarget {
	.CodeMirror {
		height: auto;
		z-index: -1;
		position: absolute;
		top: 0;
		left: 0;
		width: 100%;
		box-sizing: border-box;
		pointer-events: none;
		background: transparent;

		// Core VE default padding
		padding: 1.5em;

		// Skin specific paddings
		.skin-vector & {
			padding: 0 1rem;

			@media screen and ( min-width: 982px ) {
				.skin-vector-legacy& {
					padding: 0 1.5rem;
				}
			}
		}

		.skin-minerva &,
		.skin-monobook & {
			padding: 0;
		}
	}

	.CodeMirror-lines,
	.CodeMirror pre.CodeMirror-line,
	.CodeMirror pre.CodeMirror-line-like {
		padding: 0;
	}

	.CodeMirror-scroll {
		margin-right: 0;
		overflow: auto !important; /* stylelint-disable-line declaration-no-important */
	}

	.CodeMirror-sizer {
		border-right: 0;
	}

	.CodeMirror-selected {
		display: none;
	}

	.CodeMirror pre.cm-mw-section-1,
	.CodeMirror pre.cm-mw-section-2 {
		font-size: inherit;
		line-height: inherit;
	}

	// Ensure surfaces are using identical font rules
	.CodeMirror-code,
	.CodeMirror-code *,
	.ve-ui-mwWikitextSurface .ve-ce-paragraphNode {
		// The following are already set by mw-editfont-monospace on the parent: font-size, font-family
		line-height: 1.5em;
		word-wrap: break-word;
		// Support: Chrome<76, Firefox<69
		// Fallback for browsers which don't support break-spaces
		white-space: pre-wrap;
		// T347902
		white-space: break-spaces;
		word-break: normal;
		-webkit-hyphens: manual;
		-moz-hyphens: manual;
		-ms-hyphens: manual;
		hyphens: manual;
		-webkit-font-variant-ligatures: contextual;
		font-variant-ligatures: contextual;

		// Monospace fonts can change width when bold
		// stylelint-disable-next-line declaration-no-important
		font-weight: normal !important;
		// T252965
		line-break: initial;
	}

	.CodeMirror-gutters {
		background-color: transparent;
		border-right: 0;
	}
	.CodeMirror-linenumber {
		color: #72777d; // Codex Gray300
		padding-left: 0;
	}
	.CodeMirror-linenumber-padding {
		// Create a 12px padding between the line numbers and content.
		width: 8px;
	}
}

.ve-ce-documentNode-codeEditor-hide {
	opacity: 0.4;

	&::selection,
	& *::selection {
		background: #6da9f7 !important; /* stylelint-disable-line declaration-no-important */
	}
}

.ve-ce-documentNode-codeEditor-webkit-hide {
	-webkit-text-fill-color: transparent;
}
